@@ -1,5 +1,6 @@
 #Copyright (c) 2013 Will Harris
 # See the file license.txt for copying permission.
 
-from document import Document, Draft
-
+from document import Document
+from builder import Builder
+import drafts
