@@ -6,18 +6,23 @@ Manipulating HAL documents.
 
 import urlparse
 import itertools
-import curie
-import link
 import UserDict
 from functools import wraps
 
+import curie
+import link
+from drafts import AUTO
+from drafts import LINKS_KEY
+from drafts import EMBEDDED_KEY
+
+
 class CanonicalRels(UserDict.DictMixin, object):
     """Smart querying of link relationship types and link relationships.
 
     A ``CanonicalRels`` instance is a read-only dictionary-like object that
     provides smart retrieval and de-duplication by link relationship type. It
     is used to make access to links and embedded resources more convenient.
-    
+
     In addition to well-know link relationship types (eg. ``"self"``), keys can
     be custom link relationship types represented as URIs (eg.
     ``"http://example.com/rels/comments"``), or as URI references (eg.
@@ -39,7 +44,7 @@ class CanonicalRels(UserDict.DictMixin, object):
     """
     def __init__(self, rels, curies, base_uri, item_filter=lambda _: True):
         """Create a ``CanonicalRels`` instance.
-        
+
         Arguments:
 
         - ``rels``:        the relationships to be queried. ``rels`` should be
@@ -115,7 +120,7 @@ class CanonicalRels(UserDict.DictMixin, object):
     def __contains__(self, key):
         """Returns ``True`` if there are any link relationships for for
         ``self[key].``
-        
+
         """
         return self.canonical_key(key) in self.rels
 
@@ -135,7 +140,7 @@ class Relationships(UserDict.DictMixin, object):
     Relationships, that is links and embedded resources, are presented as a
     dictionary-like object mapping the full URI of the link relationship type
     to a list of relationships.
-    
+
     If there are both embedded resources and links for the same link relation
     type, the embedded resources will appear before the links. Otherwise,
     relationships are presented in the order they appear in their respective
@@ -144,7 +149,7 @@ class Relationships(UserDict.DictMixin, object):
     Relationships are de-duplicated by their URL, as defined by their ``self``
     link in the case of embedded resources and by their ``href`` in the case of
     links. Only the first relationship with that URL will be included.
-    
+
     """
 
     def __init__(self, links, embedded, curies, base_uri):
@@ -166,6 +171,7 @@ class Relationships(UserDict.DictMixin, object):
         rels = itertools.chain(embedded.iteritems(), links.iteritems())
 
         existing_urls = set()
+
         def item_filter(item):
             url = item.url()
             if url is not None and url in existing_urls:
@@ -173,7 +179,10 @@ class Relationships(UserDict.DictMixin, object):
             existing_urls.add(item.url())
             return True
 
-        self.canonical_rels = CanonicalRels(rels, curies, base_uri, item_filter)
+        self.canonical_rels = CanonicalRels(rels,
+                                            curies,
+                                            base_uri,
+                                            item_filter)
 
     def __getitem__(self, key):
         value = self.canonical_rels.__getitem__(key)
@@ -183,93 +192,29 @@ class Relationships(UserDict.DictMixin, object):
 
     def keys(self):
         return self.canonical_rels.keys()
-         
 
-def mutator(fn):
+
+def mutator(*cache_names):
     """Decorator for ``Document`` methods that change the document.
 
     This decorator ensures that the object's caches are kept in sync
     when changes are made.
-    
+
     """
-    @wraps(fn)
-    def deco(self, *args, **kwargs):
-        try:
-            return fn(self, *args, **kwargs)
-        finally:
-            self.prepare_cache()
+    def deco(fn):
+        @wraps(fn)
+        def _fn(self, *args, **kwargs):
+            try:
+                return fn(self, *args, **kwargs)
+            finally:
+                for cache_name in cache_names:
+                    setattr(self, cache_name, None)
+
+        return _fn
 
     return deco
 
 
-LINKS_KEY = '_links'
-EMBEDDED_KEY = '_embedded'
-
-class Draft(object):
-    """Draft-specific behaviours."""
-
-    class Draft3(object):
-        """Behaviour that is compatibile with draft 3 of the HAL spec.
-
-        CURIEs are stored in a link with the relation type 'curie'. If there are
-        multiple CURIEs, the document's JSON representation has a JSON array in
-        ``_links.curie``, but if there is only one CURIE, ``_links.curie`` is a
-        JSON object.
-
-        See http://tools.ietf.org/html/draft-kelly-json-hal-03.
-        """
-
-        curies_rel = 'curie'
-
-        def detect(self, obj):
-            return self
-
-        def set_curie(self, doc, name, href):
-            doc.add_link(self.curies_rel, href, name=name, templated=True)
-
-    class Draft4(object):
-        """Behaviour that is compatibile with draft 4 of the HAL spec.
-
-        CURIEs are stored in a link with the relation type 'curies'. If there
-        are one or more CURIEs, the document's JSON representation has a JSON
-        array in ``_links.curies``.
-
-        See http://tools.ietf.org/html/draft-kelly-json-hal-04.
-        """
-
-        curies_rel = 'curies'
-
-        def detect(self, obj):
-            return self
-
-        def set_curie(self, doc, name, href):
-            # CURIE links should always be in an array, even if there is only
-            # one.
-            doc.o.setdefault(LINKS_KEY, {}).setdefault(self.curies_rel, [])
-            doc.add_link(self.curies_rel, href, name=name, templated=True)
-
-    class DraftAuto(object):
-        """Behaviour for documents that automatically detect draft version.
-        
-        When created with an existing JSON object, the document guesses the
-        draft version based on the presence of a link with a relation type of
-        'curie' or 'curies'.
-        """
-
-        def detect(self, obj):
-            links = obj.get(LINKS_KEY, {})
-
-            for draft in [Draft.DRAFT_4, Draft.DRAFT_3]:
-                if draft.curies_rel in links:
-                    return draft
-
-            return Draft.DRAFT_4
-
-    DRAFT_3 = Draft3()
-    DRAFT_4 = Draft4()
-    LATEST = DRAFT_4
-    AUTO = DraftAuto()
-
 class Document(object):
     """Represents the document for a HAL resource.
 
@@ -286,82 +231,117 @@ class Document(object):
                       excluding ``_links`` and ``_embedded``. ``properties``
                       should be treated as read-only.
     - ``links``: ``dict`` containing the document's links, excluding
-                 ``curies``. Each link relationship type is mapped to a ``Link``
-                 instance or a list of ``Link`` instances. ``links`` should be
-                 treated as read-only.
+                 ``curies``. Each link relationship type is mapped to a
+                 ``Link`` instance or a list of ``Link`` instances. ``links``
+                 should be treated as read-only.
     - ``embedded``: dictionary containing the document's embedded resources.
                     Each link relationship type is mapped to a ``Document``
                     instance.
     - ``rels``: a ``Relationships`` instance holding a merged view of the
                 relationships from the document.
     - ``draft``: a ``Draft`` instance that selects the version of the spec to
-                 which the document should conform. Defaults to ``Draft.AUTO``.
+                 which the document should conform. Defaults to
+                 ``drafts.AUTO``.
 
     """
-    def __init__(self, o, base_uri, parent_curies=None, draft=Draft.AUTO):
+    def __init__(self, o, base_uri, parent_curies=None, draft=AUTO):
+        self.prepare_cache()
         self.o = o
         self.base_uri = base_uri
         self.parent_curies = parent_curies
         self.draft = draft.detect(o)
-        self.prepare_cache()
 
     RESERVED_ATTRIBUTE_NAMES = (LINKS_KEY, EMBEDDED_KEY)
 
-    def prepare_cache(self):
-        def properties_cache():
-            properties = dict(self.o)
-            for name in self.RESERVED_ATTRIBUTE_NAMES:
-                properties[name] = None
-                del properties[name]
-            return properties
+    def properties_cache(self):
+        properties = dict(self.o)
+        for name in self.RESERVED_ATTRIBUTE_NAMES:
+            properties[name] = None
+            del properties[name]
+        return properties
 
-        def links_cache():
-            links = {}
+    def links_cache(self):
+        links = {}
 
-            links_json = self.o.get(LINKS_KEY, {})
+        links_json = self.o.get(LINKS_KEY, {})
 
-            for key, value in links_json.iteritems():
-                if key == self.draft.curies_rel:
-                    continue
-                links[key] = link.Link.from_object(value, self.base_uri)
+        for key, value in links_json.iteritems():
+            if key == self.draft.curies_rel:
+                continue
+            links[key] = link.Link.from_object(value, self.base_uri)
 
-            return CanonicalRels(links, self.curies, self.base_uri)
+        return CanonicalRels(links, self.curies, self.base_uri)
 
-        def load_curie_collection():
-            result = curie.CurieCollection()
-            if self.parent_curies is not None:
-                result.update(self.parent_curies)
+    def curies_cache(self):
+        result = curie.CurieCollection()
+        if self.parent_curies is not None:
+            result.update(self.parent_curies)
 
-            links_json = self.o.get('_links', {})
-            curies_json = links_json.get(self.draft.curies_rel)
-
-            if not curies_json:
-                return result
-
-            curies = link.Link.from_object(curies_json, self.base_uri)
-
-            if not isinstance(curies, list):
-                curies = [curies]
-
-            for curie_link in curies:
-                result[curie_link.name] = curie_link
+        links_json = self.o.get('_links', {})
+        curies_json = links_json.get(self.draft.curies_rel)
 
+        if not curies_json:
             return result
 
-        def embedded_cache():
-            embedded = {}
-            for key, value in self.o.get(EMBEDDED_KEY, {}).iteritems():
-                embedded[key] = self.from_object(value,
-                                                 self.base_uri,
-                                                 self.curies)
-            return CanonicalRels(embedded, self.curies, self.base_uri)
+        curies = link.Link.from_object(curies_json, self.base_uri)
 
-        self.properties = properties_cache()
-        self.curies = load_curie_collection()
-        self.links = links_cache()
-        self.embedded = embedded_cache()
-        self.rels = Relationships(self.links, self.embedded, self.curies,
-                                  self.base_uri)
+        if not isinstance(curies, list):
+            curies = [curies]
+
+        for curie_link in curies:
+            result[curie_link.name] = curie_link
+
+        return result
+
+    def embedded_cache(self):
+        embedded = {}
+        for key, value in self.o.get(EMBEDDED_KEY, {}).iteritems():
+            embedded[key] = self.from_object(value,
+                                             self.base_uri,
+                                             self.curies)
+
+        return CanonicalRels(embedded, self.curies, self.base_uri)
+
+    def rels_cache(self):
+        return Relationships(self.links, self.embedded, self.curies,
+                             self.base_uri)
+
+    def prepare_cache(self):
+        self._properties_cache = None
+        self._curies_cache = None
+        self._links_cache = None
+        self._embedded_cache = None
+        self._rels_cache = None
+
+    @property
+    def properties(self):
+        if self._properties_cache is None:
+            self._properties_cache = self.properties_cache()
+        return self._properties_cache
+
+    @property
+    def curies(self):
+        if self._curies_cache is None:
+            self._curies_cache = self.curies_cache()
+        return self._curies_cache
+
+    @property
+    def links(self):
+        if self._links_cache is None:
+            self._links_cache = self.links_cache()
+        return self._links_cache
+
+    @property
+    def embedded(self):
+        if self._embedded_cache is None:
+            self._embedded_cache = self.embedded_cache()
+        return self._embedded_cache
+
+    @property
+    def rels(self):
+        if self._rels_cache is None:
+            self._rels_cache = self.rels_cache()
+        return self._rels_cache
 
     def url(self):
         """Returns the URL for the resource based on the ``self`` link.
@@ -369,7 +349,7 @@ class Document(object):
         This method returns the ``href`` of the document's ``self`` link if it
         has one, or ``None`` if the document lacks a ``self`` link, or the
         ``href`` of the document's first ``self`` link if it has more than one.
-        
+
         """
         if not 'self' in self.links:
             return None
@@ -402,7 +382,7 @@ class Document(object):
         """Returns a ``Link`` to the resource."""
         return self.links['self']
 
-    @mutator
+    @mutator('_properties_cache')
     def set_property(self, key, value):
         """Set a property on the document.
 
@@ -422,9 +402,9 @@ class Document(object):
             return
         self.o[key] = value
 
-    @mutator
+    @mutator('_properties_cache')
     def delete_property(self, key):
-        """Remove an property from the document.
+        """Remove a property from the document.
 
         Calling code should use this method to remove properties on the
         document instead of modifying ``properties`` directly.
@@ -441,13 +421,13 @@ class Document(object):
         """Retuns a new link relative to this resource."""
         return link.Link(dict(href=href, **kwargs), self.base_uri)
 
-    @mutator
-    def add_link(self, rel, target, **kwargs):
+    @mutator('_links_cache')
+    def add_link(self, rel, target, wrap=False, **kwargs):
         """Adds a link to the document.
 
         Calling code should use this method to add links instead of
         modifying ``links`` directly.
-        
+
         This method adds a link to the given ``target`` to the document with
         the given ``rel``. If one or more links are already present for that
         link relationship type, the new link will be added to the existing
@@ -462,6 +442,9 @@ class Document(object):
         If ``target`` is a ``Document`` object, ``target``'s ``self`` link is
         added to this document and the keyword arguments are ignored.
 
+        If ``target`` is a ``Builder`` object, ``target``'s ``self`` link is
+        added to this document and the keyword arguments are ignored.
+
         Arguments:
 
         - ``rel``: a string specifying the link relationship type of the link.
@@ -469,6 +452,9 @@ class Document(object):
           (http://www.iana.org/assignments/link-relations/link-relations.xml),
           a full URI, or a CURIE.
         - ``target``: the destination of the link.
+        - ``wrap``: Defaults to False, but if True, specifies that the link
+          object should be initally wrapped in a JSON array even if it is the
+          first link for the given ``rel``.
 
         """
         if hasattr(target, 'as_link'):
@@ -477,11 +463,14 @@ class Document(object):
             link = self.link(target, **kwargs)
 
         links = self.o.setdefault(LINKS_KEY, {})
-        
+
         new_link = link.as_object()
         collected_links = CanonicalRels(links, self.curies, self.base_uri)
         if rel not in collected_links:
-            links[rel] = new_link
+            if wrap:
+                links[rel] = [new_link]
+            else:
+                links[rel] = new_link
             return
 
         original_rel = collected_links.original_key(rel)
@@ -492,7 +481,7 @@ class Document(object):
         else:
             links[original_rel] = [current_links, new_link]
 
-    @mutator
+    @mutator('_links_cache')
     def delete_link(self, rel=None, href=lambda _: True):
         """Deletes links from the document.
 
@@ -551,8 +540,7 @@ class Document(object):
             del self.o[LINKS_KEY]
 
     @classmethod
-    def from_object(cls, o, base_uri=None, parent_curies=None,
-                    draft=Draft.AUTO):
+    def from_object(cls, o, base_uri=None, parent_curies=None, draft=AUTO):
         """Returns a new ``Document`` based on a JSON object or array.
 
         Arguments:
@@ -563,11 +551,11 @@ class Document(object):
                                relative URLs in the document.
         - ``parent_curies``: optional ``CurieCollection`` instance holding the
                              CURIEs of the parent document in which the new
-                             document is to be embedded. Calling code should not
-                             normally provide this argument.
+                             document is to be embedded. Calling code should
+                             not normally provide this argument.
         - ``draft``: a ``Draft`` instance that selects the version of the spec
                      to which the document should conform. Defaults to
-                     ``Draft.AUTO``.
+                     ``drafts.AUTO``.
 
         """
 
@@ -581,7 +569,7 @@ class Document(object):
         return cls(o, base_uri, parent_curies, draft)
 
     @classmethod
-    def empty(cls, base_uri=None, draft=Draft.AUTO):
+    def empty(cls, base_uri=None, draft=AUTO):
         """Returns an empty ``Document``.
 
         Arguments:
@@ -590,12 +578,12 @@ class Document(object):
                                relative URLs in the document.
         - ``draft``: a ``Draft`` instance that selects the version of the spec
                      to which the document should conform. Defaults to
-                     ``Draft.AUTO``.
+                     ``drafts.AUTO``.
         """
         return cls.from_object({}, base_uri=base_uri, draft=draft)
 
-    @mutator
-    def embed(self, rel, other):
+    @mutator('_embedded_cache')
+    def embed(self, rel, other, wrap=False):
         """Embeds a document inside this document.
 
         Arguments:
@@ -606,33 +594,57 @@ class Document(object):
           (http://www.iana.org/assignments/link-relations/link-relations.xml),
           a full URI, or a CURIE.
         - ``other``: a ``Document`` instance that will be embedded in this
-          document.
+          document. If ``other`` is identical to this document, this method
+          will silently fail.
+        - ``wrap``: Defaults to False, but if True, specifies that the embedded
+          resource object should be initally wrapped in a JSON array even if it
+          is the first embedded resource for the given ``rel``.
 
         Calling code should use this method to add embedded resources instead
         of modifying ``embedded`` directly.
-        
+
         This method embeds the given document in this document with the given
         ``rel``. If one or more documents have already been embedded for that
         ``rel``, the new document will be embedded in addition to those
         documents.
 
         """
-        embedded = self.o.setdefault(EMBEDDED_KEY, {})
-        collected_embedded = CanonicalRels(embedded, self.curies, self.base_uri)
 
-        if rel not in collected_embedded:
-            embedded[rel] = other.as_object()
+        if other == self:
             return
 
-        original_rel = collected_embedded.original_key(rel)
+        embedded = self.o.setdefault(EMBEDDED_KEY, {})
+        collected_embedded = CanonicalRels(embedded,
+                                           self.curies,
+                                           self.base_uri)
 
-        current_embedded = embedded[original_rel]
-        if isinstance(current_embedded, list):
-            current_embedded.append(other.as_object())
+        if rel not in collected_embedded:
+            if wrap:
+                embedded[rel] = [other.as_object()]
+            else:
+                embedded[rel] = other.as_object()
         else:
-            embedded[original_rel] = [current_embedded, other.as_object()]
+            original_rel = collected_embedded.original_key(rel)
 
-    @mutator
+            current_embedded = embedded[original_rel]
+            if isinstance(current_embedded, list):
+                current_embedded.append(other.as_object())
+            else:
+                embedded[original_rel] = [current_embedded, other.as_object()]
+
+        if not self.draft.automatic_link:
+            return
+
+        url = other.url()
+        if not url:
+            return
+
+        if url in (link.url() for link in self.links.get(rel, [])):
+            return
+
+        self.add_link(rel, other, wrap=wrap)
+
+    @mutator('_embedded_cache')
     def delete_embedded(self, rel=None, href=lambda _: True):
         """Removes an embedded resource from this document.
 
@@ -701,6 +713,7 @@ class Document(object):
         if not self.o[EMBEDDED_KEY]:
             del self.o[EMBEDDED_KEY]
 
+    @mutator('_curies_cache')
     def set_curie(self, name, href):
         """Sets a CURIE.
 
@@ -711,7 +724,7 @@ class Document(object):
 
         self.draft.set_curie(self, name, href)
 
-    @mutator
+    @mutator('_curies_cache')
     def drop_curie(self, name):
         """Removes a CURIE.
 
@@ -722,12 +735,12 @@ class Document(object):
         if isinstance(curies, dict) and curies['name'] == name:
             del self.o[LINKS_KEY][self.draft.curies_rel]
             return
-        
+
         for i, curie in enumerate(curies):
             if curie['name'] == name:
                 del curies[i]
                 break
-                
+
             continue
 
     def __iter__(self):
@@ -736,11 +749,8 @@ class Document(object):
     def __eq__(self, other):
         if not isinstance(other, Document):
             return False
-        
+
         return self.as_object() == other.as_object()
 
     def __repr__(self):
         return "<Document %r>" % self.url()
-
-
-
